'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (ref) {
  var t = ref.types;

  var id = function (name) { return t.identifier(name); };

  function pipe(fnName) {
    var args = [], len = arguments.length - 1;
    while ( len-- > 0 ) args[ len ] = arguments[ len + 1 ];

    return t.callExpression(id(fnName), args);
  }

  function importAs(name, as, from) {
    return t.importDeclaration([t.importSpecifier(t.identifier(name), t.identifier(as))], t.stringLiteral(from));
  }

  var visitor = {
    Program: {
      /* import our function if not defined */
      exit: function exit(path, ref) {
        var opts = ref.opts;

        var name = opts.name; if ( name === void 0 ) name = 'pipe';
        var as = opts.as; if ( as === void 0 ) as = 'pipe';
        var from = opts.from;

        if (!from) return;

        if (!path.scope.hasBinding(as)) {
          path.unshiftContainer('body', importAs(name, as, from));
        }
      }
    },

    CallExpression: {
      /* transform multiple pipe calls into as few pipe calls as possible */
      exit: function exit$1(path, ref) {
        var opts = ref.opts;

        var as = opts.as; if ( as === void 0 ) as = 'pipe';

        if (isPipeCall(as, path.node)) {
          var ref$1 = path.node.arguments;
          var left = ref$1[0];
          var right = ref$1[1];
          if (!left || !right) return;

          if (isPipeCall(as, left)) {
            path.replaceWith(pipe.apply(void 0, [ as ].concat( left.arguments, [right] )));
          }
        }
      }
    },

    BinaryExpression: {
      /* transform >> and << into calls to pipe() */
      enter: function enter(path, ref) {
        var opts = ref.opts;

        if (!hasDirective(path)) return;
        var as = opts.as; if ( as === void 0 ) as = 'pipe';

        if (isPipe(path)) {
          var ref$1 = side(path);
          var left = ref$1[0];
          var right = ref$1[1];
          if (!left.node || !right.node) return;

          path.replaceWith(pipe(as, left.node, right.node));
        }

        if (isAntiPipe(path)) {
          var ref$2 = side(path);
          var left$1 = ref$2[0];
          var right$1 = ref$2[1];
          if (!left$1.node || !right$1.node) return;

          path.replaceWith(pipe(as, right$1.node, left$1.node));
        }

        if (isTriplePipe(path)) {
          var ref$3 = side(path);
          var left$2 = ref$3[0];
          var right$2 = ref$3[1];
          if (!left$2.node || !right$2.node) return;

          if (t.isIdentifier(right$2)) {
            path.replaceWith(pipe(right$2.node.name, left$2.node));
          }

          if (t.isCallExpression(right$2)) {
            path.replaceWith(pipe.apply(void 0, [ right$2.node.callee.name ].concat( right$2.node.arguments, [left$2.node] )));
          }
        }
      }
    }
  };

  return { visitor: visitor };
};

var DIRECTIVE = 'use fc';
var LEFTPIPE = '>>';
var RIGHTPIPE = '<<';
var TRIPLEPIPE = '>>>';

/* check to see if scope has directive "use pipe" */
function hasDirective(path) {
  var matched = false;

  path.findParent(function (ref) {
    var node = ref.node;

    node.directives && node.directives.some(function (ref) {
      var value = ref.value;

      matched = value.value === DIRECTIVE;
    });
  });

  return matched;
}

/* check if an ast node is a function call pipe */
function isPipeCall(as, node) {
  return node.callee && node.callee.name === as;
}

/* checks if an BinaryExpression is an pipe call */
function isPipe(path) {
  return path.node.operator === LEFTPIPE;
}

/* checks if an BinaryExpression is a right pipe call */
function isAntiPipe(path) {
  return path.node.operator === RIGHTPIPE;
}

function isTriplePipe(path) {
  return path.node.operator === TRIPLEPIPE;
}

/* gets the nodes to the left and right of a path */
function side(path) {
  return [leftOf(path), rightOf(path)];
}

/* gets the path left of a given path */
function leftOf(path) {
  return path.get('left');
}

/* get the path right of a given path */
function rightOf(path) {
  return path.get('right');
}